import numpy as np
import torch

"""
This module contains methods for trasforming an input signal
in a BVP signal using a rPPG method (see pyVHR.BVP.methods).
"""

def signals_to_bvps_torch(sig, torch_method, params={}):
    """
    Transform an input RGB signal in a BVP signal using a rPPG 
    method (see pyVHR.BVP.methods).
    This method must use Torch and execute on GPU/CPU.
    You can pass also non-RGB signal but the method used must handle its shape.

    Args:
        sig (float32 ndarray): RGB Signal as float32 ndarray with shape  [num_estimators, rgb_channels, num_frames].
            You can pass also a generic signal but the method used must handle its shape and type.
        torch_method: a method that comply with the fucntion signature documented 
            in pyVHR.BVP.methods. This method must use Torch.
        params (dict): dictionary of usefull parameters that will be passed to the method.
    
    Returns:
        float32 ndarray: BVP signal as float32 ndarray with shape [num_estimators, num_frames].
    """
    if sig.shape[0] == 0:
        return np.zeros((0, sig.shape[2]), dtype=sig.dtype)
    cpu_sig = torch.from_numpy(sig)
    if len(params) > 0:
        bvps = torch_method(cpu_sig, **params)
    else:
        bvps = torch_method(cpu_sig)
    return bvps.numpy()


def signals_to_bvps_cpu(sig, cpu_method, params={}):
    """
    Transform an input RGB signal in a BVP signal using a rPPG 
    method (see pyVHR.BVP.methods).
    This method must use and execute on CPU.
    You can pass also non-RGB signal but the method used must handle its shape.

    Args:
        sig (float32 ndarray): RGB Signal as float32 ndarray with shape  [num_estimators, rgb_channels, num_frames].
            You can pass also a generic signal but the method used must handle its shape and type.
        cpu_method: a method that comply with the fucntion signature documented 
            in pyVHR.BVP.methods. This method must use Numpy.
        params (dict): dictionary of usefull parameters that will be passed to the method.
    
    Returns:
        float32 ndarray: BVP signal as float32 ndarray with shape [num_estimators, num_frames].
    """
    if sig.shape[0] == 0:
        return np.zeros((0, sig.shape[2]), dtype=sig.dtype)
    cpu_sig = np.array(sig)
    if len(params) > 0:
        bvps = cpu_method(cpu_sig, **params)
    else:
        bvps = cpu_method(cpu_sig)
    return bvps


def RGB_sig_to_BVP(windowed_sig, fps, device_type=None, method=None, params={}):
    """
    Transform an input RGB windowed signal in a BVP windowed signal using a rPPG method (see pyVHR.BVP.methods).
    You can pass also non-RGB signal but the method used must handle its shape.

    Args:
        windowed_sig (list): RGB windowed signal as a list of length num_windows of np.ndarray with shape [num_estimators, rgb_channels, num_frames].
        fps (float): frames per seconds. You can pass also a generic signal but the method used must handle its shape and type.
        device_type (str): the chosen rPPG method run on CPU ('cpu', 'torch').
        method: a method that comply with the fucntion signature documented 
            in pyVHR.BVP.methods. This method must use Numpy if the 'device_type' is 'cpu', Torch if the 'device_type' is 'torch', and Cupy 
            if the 'device_type' is 'cuda'.
        params(dict): dictionary of usefull parameters that will be passed to the method. If the method needs fps you can set {'fps':'adaptive'}
            in the dictionary to use the 'fps' input variable.

    Returns:
        a list of lenght num_windows of BVP signals as np.ndarray with shape [num_estimators, num_frames];
        if no BVP can be found in a window, then the np.ndarray has num_estimators == 0.
    """
    if device_type != 'cpu' and device_type != 'torch':
        print("[ERROR]: invalid device_type!")
        return []

    if 'fps' in params and params['fps'] == 'adaptive':
        params['fps'] = np.float32(fps)

    bvps = []
    for sig in windowed_sig:
        copy_signal = np.copy(sig)
        bvp = np.zeros((0, 1), dtype=np.float32)
        if device_type == 'cpu':
            bvp = signals_to_bvps_cpu(
                copy_signal, method, params)
        elif device_type == 'torch':
            bvp = signals_to_bvps_torch(
                copy_signal, method, params)
        bvps.append(bvp)
    return bvps


def concatenate_BVPs(list_of_BVPs):
    """
    Join a list of windowed BVPs. There must be the same number of windows, and each one must have the same number of frames.

    Args:
        list_of_BVPs (list): a list of windowed BVPs each one defined as a list of lenght num_windows of BVP signals as np.ndarray
            with shape [num_estimators, num_frames]. Remember that concatenation is possible only if 'num_frames'
            is the same for each BVP of the nth window.
            
    Returns:
        a list of lenght num_windows of BVP signals as np.ndarray with shape [total_num_estimators, num_frames];
        if no BVP can be found in a window, then the np.ndarray has total_num_estimators == 0. 
        For example: given the BVP windows of shape [10, 200] and [20,200], the concatenated window will have shape [30,200];
        If an exception is thrown for any reason, the function returns 0.
    """
    if len(list_of_BVPs) <= 1:
        return 0
    
    first_window_len = len(list_of_BVPs[0])

    for item in list_of_BVPs:
        if first_window_len != len(item):
            return 0

    try:
        concatenated_BVPs = []
        for i in range(first_window_len):
            concatenated_BVPs.append(np.concatenate(tuple([e[i] for e in list_of_BVPs]), axis=0))
        return concatenated_BVPs
    except ValueError as e:
        print(e)
        return 0


from numpy import abs, append, arange, arctan2, argsort, array, concatenate, \
    cos, diag, dot, eye, float32,float64, matrix, multiply, ndarray, newaxis, \
    sign, sin, sqrt, zeros
import numpy as np

def jadeR(X, m=None, verbose=True):
    """
    Blind separation of real signals with JADE.
    jadeR implements JADE, an Independent Component Analysis (ICA) algorithm
    developed by Jean-Francois Cardoso. See http://www.tsi.enst.fr/~cardoso/guidesepsou.html , and papers cited
    at the end of the source file.
    
    Translated into NumPy from the original Matlab Version 1.8 (May 2005) by
    Gabriel Beckers, http://gbeckers.nl .

    Parameters:
        X -- an nxT data matrix (n sensors, T samples). May be a numpy array or
             matrix.
        m -- output matrix B has size mxn so that only m sources are
             extracted.  This is done by restricting the operation of jadeR
             to the m first principal components. Defaults to None, in which
             case m=n.
        verbose -- print info on progress. Default is True.

    Returns:
        An m*n matrix B (NumPy matrix type), such that Y=B*X are separated
        sources extracted from the n*T data matrix X. If m is omitted, B is a
        square n*n matrix (as many sources as sensors). The rows of B are
        ordered such that the columns of pinv(B) are in order of decreasing
        norm; this has the effect that the `most energetically significant`
        components appear first in the rows of Y=B*X.

    Quick notes (more at the end of this file):
    o This code is for REAL-valued signals.  A MATLAB implementation of JADE
    for both real and complex signals is also available from
    http://sig.enst.fr/~cardoso/stuff.html
    o This algorithm differs from the first released implementations of
    JADE in that it has been optimized to deal more efficiently
    1) with real signals (as opposed to complex)
    2) with the case when the ICA model does not necessarily hold.
    o There is a practical limit to the number of independent
    components that can be extracted with this implementation.  Note
    that the first step of JADE amounts to a PCA with dimensionality
    reduction from n to m (which defaults to n).  In practice m
    cannot be `very large` (more than 40, 50, 60... depending on
    available memory)
    o See more notes, references and revision history at the end of
    this file and more stuff on the WEB
    http://sig.enst.fr/~cardoso/stuff.html
    o For more info on NumPy translation, see the end of this file.
    o This code is supposed to do a good job!  Please report any
    problem relating to the NumPY code gabriel@gbeckers.nl

    Copyright original Matlab code : Jean-Francois Cardoso <cardoso@sig.enst.fr>
    Copyright Numpy translation : Gabriel Beckers <gabriel@gbeckers.nl>
    """

    # GB: we do some checking of the input arguments and copy data to new
    # variables to avoid messing with the original input. We also require double
    # precision (float64) and a numpy matrix type for X.

    assert isinstance(X, ndarray),\
        "X (input data matrix) is of the wrong type (%s)" % type(X)
    origtype = X.dtype # remember to return matrix B of the same type
    X = matrix(X.astype(float64))
    assert X.ndim == 2, "X has %d dimensions, should be 2" % X.ndim
    assert (verbose == True) or (verbose == False), \
        "verbose parameter should be either True or False"

    [n,T] = X.shape # GB: n is number of input signals, T is number of samples

    if m==None:
        m=n   # Number of sources defaults to # of sensors
    assert m<=n,\
        "jade -> Do not ask more sources (%d) than sensors (%d )here!!!" % (m,n)

    if verbose:
        print ("jade -> Looking for %d sources" % m)
        print ( "jade -> Removing the mean value")
    X -= X.mean(1)

    # whitening & projection onto signal subspace
    # ===========================================
    if verbose:
        print ("jade -> Whitening the data")
    [D,U] = np.linalg.eig((X * X.T) / float(T)) # An eigen basis for the sample covariance matrix
    k = D.argsort()
    Ds = D[k] # Sort by increasing variances
    PCs = arange(n-1, n-m-1, -1)    # The m most significant princip. comp. by decreasing variance

    # --- PCA  ----------------------------------------------------------
    B = U[:,k[PCs]].T    # % At this stage, B does the PCA on m components

    # --- Scaling  ------------------------------------------------------
    scales = sqrt(Ds[PCs]) # The scales of the principal components .
    B = diag(1./scales) * B  # Now, B does PCA followed by a rescaling = sphering
    #B[-1,:] = -B[-1,:] # GB: to make it compatible with octave
    # --- Sphering ------------------------------------------------------
    X = B * X # %% We have done the easy part: B is a whitening matrix and X is white.

    del U, D, Ds, k, PCs, scales

    # NOTE: At this stage, X is a PCA analysis in m components of the real data, except that
    # all its entries now have unit variance.  Any further rotation of X will preserve the
    # property that X is a vector of uncorrelated components.  It remains to find the
    # rotation matrix such that the entries of X are not only uncorrelated but also `as
    # independent as possible".  This independence is measured by correlations of order
    # higher than 2.  We have defined such a measure of independence which
    #   1) is a reasonable approximation of the mutual information
    #   2) can be optimized by a `fast algorithm"
    # This measure of independence also corresponds to the `diagonality" of a set of
    # cumulant matrices.  The code below finds the `missing rotation " as the matrix which
    # best diagonalizes a particular set of cumulant matrices.


    # Estimation of the cumulant matrices.
    # ====================================
    if verbose:
        print ("jade -> Estimating cumulant matrices")

    # Reshaping of the data, hoping to speed up things a little bit...
    X = X.T
    dimsymm = int((m * ( m + 1)) / 2) # Dim. of the space of real symm matrices
    nbcm = dimsymm  # number of cumulant matrices
    CM = matrix(zeros([m,m*nbcm], dtype=float64)) # Storage for cumulant matrices
    R = matrix(eye(m, dtype=float64))
    Qij = matrix(zeros([m,m], dtype=float64)) # Temp for a cum. matrix
    Xim = zeros(m, dtype=float64) # Temp
    Xijm = zeros(m, dtype=float64) # Temp
    #Uns = numpy.ones([1,m], dtype=numpy.uint32)    # for convenience
    # GB: we don't translate that one because NumPy doesn't need Tony's rule

    # I am using a symmetry trick to save storage.  I should write a short note one of these
    # days explaining what is going on here.
    Range = arange(m) # will index the columns of CM where to store the cum. mats.

    for im in range(m):
        Xim = X[:,im]
        Xijm = multiply(Xim, Xim)
        # Note to myself: the -R on next line can be removed: it does not affect
        # the joint diagonalization criterion
        Qij = multiply(Xijm, X).T * X / float(T)\
            - R - 2 * dot(R[:,im], R[:,im].T)
        CM[:,Range] = Qij
        Range = Range  + m
        for jm in range(im):
            Xijm = multiply(Xim, X[:,jm])
            Qij = sqrt(2) * multiply(Xijm, X).T * X / float(T) \
                - R[:,im] * R[:,jm].T - R[:,jm] * R[:,im].T
            CM[:,Range] = Qij
            Range = Range + m

    # Now we have nbcm = m(m+1)/2 cumulants matrices stored in a big m x m*nbcm array.

    V = matrix(eye(m, dtype=float64))

    Diag = zeros(m, dtype=float64)
    On = 0.0
    Range = arange(m)
    for im in range(nbcm):
        Diag = diag(CM[:,Range])
        On = On + (Diag*Diag).sum(axis=0)
        Range = Range + m
    Off = (multiply(CM,CM).sum(axis=0)).sum(axis=0) - On

    seuil = 1.0e-6 / sqrt(T) # % A statistically scaled threshold on `small" angles
    encore = True
    sweep = 0 # % sweep number
    updates = 0 # % Total number of rotations
    upds = 0 # % Number of rotations in a given seep
    g = zeros([2,nbcm], dtype=float64)
    gg = zeros([2,2], dtype=float64)
    G = zeros([2,2], dtype=float64)
    c = 0
    s = 0
    ton = 0
    toff = 0
    theta = 0
    Gain = 0

    # Joint diagonalization proper

    if verbose:
        print ( "jade -> Contrast optimization by joint diagonalization")

    while encore:
        encore = False
        if verbose:
            print("jade -> Sweep #%3d" % sweep)
        sweep = sweep + 1
        upds  = 0
        Vkeep = V

        for p in range(m-1):
            for q in range(p+1, m):

                Ip = arange(p, m*nbcm, m)
                Iq = arange(q, m*nbcm, m)

                # computation of Givens angle
                g = concatenate([CM[p,Ip] - CM[q,Iq], CM[p,Iq] + CM[q,Ip]])
                gg = dot(g, g.T)
                ton = gg[0,0] - gg[1,1]
                toff = gg[0,1] + gg[1,0]
                theta = 0.5 * arctan2(toff, ton + sqrt(ton * ton + toff * toff))
                Gain = (sqrt(ton * ton + toff * toff) - ton) / 4.0

                # Givens update
                if abs(theta) > seuil:
                    encore = True
                    upds = upds + 1
                    c = cos(theta)
                    s = sin(theta)
                    G = matrix([[c, -s] , [s, c] ])
                    pair = array([p,q])
                    V[:,pair] = V[:,pair] * G
                    CM[pair,:] = G.T * CM[pair,:]
                    CM[:,concatenate([Ip,Iq])] = \
                        append( c*CM[:,Ip]+s*CM[:,Iq], -s*CM[:,Ip]+c*CM[:,Iq], \
                               axis=1)
                    On = On + Gain
                    Off = Off - Gain

        if verbose:
            print ( "completed in %d rotations" % upds)
        updates = updates + upds
    if verbose:
        print ("jade -> Total of %d Givens rotations" % updates)

    # A separating matrix
    # ===================

    B = V.T * B

    # Permute the rows of the separating matrix B to get the most energetic components first.
    # Here the **signals** are normalized to unit variance.  Therefore, the sort is
    # according to the norm of the columns of A = pinv(B)

    if verbose:
        print("jade -> Sorting the components")

    A = np.linalg.pinv(B)
    keys =  array(argsort(multiply(A,A).sum(axis=0)[0]))[0]
    B = B[keys,:]
    B = B[::-1,:]     # % Is this smart ?


    if verbose:
        print ("jade -> Fixing the signs")
    b = B[:,0]
    signs = array(sign(sign(b)+0.1).T)[0] # just a trick to deal with sign=0
    B = diag(signs) * B

    return B.astype(origtype)



#!/usr/bin/python
# -*- encoding: utf-8 -*-

from logger import setup_logger
from model import BiSeNet

import torch

import os
import os.path as osp
import numpy as np
from PIL import Image
import torchvision.transforms as transforms
import cv2

def vis_parsing_maps(im, parsing_anno, stride, save_im=False, save_path='vis_results/parsing_map_on_im.jpg'):
    # Colors for all 20 parts
    part_colors = [[255, 0, 0], [255, 85, 0], [255, 170, 0],
                   [255, 0, 85], [255, 0, 170],
                   [0, 255, 0], [85, 255, 0], [170, 255, 0],
                   [0, 255, 85], [0, 255, 170],
                   [0, 0, 255], [85, 0, 255], [170, 0, 255],
                   [0, 85, 255], [0, 170, 255],
                   [255, 255, 0], [255, 255, 85], [255, 255, 170],
                   [255, 0, 255], [255, 85, 255], [255, 170, 255],
                   [0, 255, 255], [85, 255, 255], [170, 255, 255]]

    im = np.array(im)
    vis_im = im.copy().astype(np.uint8)
    vis_parsing_anno = parsing_anno.copy().astype(np.uint8)
    vis_parsing_anno = cv2.resize(vis_parsing_anno, None, fx=stride, fy=stride, interpolation=cv2.INTER_NEAREST)
    vis_parsing_anno_color = np.zeros((vis_parsing_anno.shape[0], vis_parsing_anno.shape[1], 3)) + 255

    num_of_class = np.max(vis_parsing_anno)

    for pi in range(1, num_of_class + 1):
        index = np.where(vis_parsing_anno == pi)
        vis_parsing_anno_color[index[0], index[1], :] = part_colors[pi]

    vis_parsing_anno_color = vis_parsing_anno_color.astype(np.uint8)
    # print(vis_parsing_anno_color.shape, vis_im.shape)
    vis_im = cv2.addWeighted(cv2.cvtColor(vis_im, cv2.COLOR_RGB2BGR), 0.4, vis_parsing_anno_color, 0.6, 0)

    # Save result or not
    if save_im:
        cv2.imwrite(save_path[:-4] +'.png', vis_parsing_anno)
        cv2.imwrite(save_path, vis_im, [int(cv2.IMWRITE_JPEG_QUALITY), 100])

    # return vis_im

def evaluate(respth='./res/test_res', dspth='./data', cp='model_final_diss.pth'):

    if not os.path.exists(respth):
        os.makedirs(respth)

    n_classes = 19
    net = BiSeNet(n_classes=n_classes)
    net.cuda()
    save_pth = osp.join('res/cp', cp)
    net.load_state_dict(torch.load(save_pth))
    net.eval()

    to_tensor = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
    ])
    with torch.no_grad():
        for image_path in os.listdir(dspth):
            img = Image.open(osp.join(dspth, image_path))
            image = img.resize((512, 512), Image.BILINEAR)
            img = to_tensor(image)
            img = torch.unsqueeze(img, 0)
            img = img.cuda()
            out = net(img)[0]
            parsing = out.squeeze(0).cpu().numpy().argmax(0)
            # print(parsing)
            print(np.unique(parsing))

            vis_parsing_maps(image, parsing, stride=1, save_im=True, save_path=osp.join(respth, image_path))







if __name__ == "__main__":
    evaluate(dspth='/home/zll/data/CelebAMask-HQ/test-img', cp='79999_iter.pth')





#!/usr/bin/python
# -*- encoding: utf-8 -*-

import torch
from torch.utils.data import Dataset
import torchvision.transforms as transforms

import os.path as osp
import os
from PIL import Image
import numpy as np
import json
import cv2

from transform import *



class FaceMask(Dataset):
    def __init__(self, rootpth, cropsize=(640, 480), mode='train', *args, **kwargs):
        super(FaceMask, self).__init__(*args, **kwargs)
        assert mode in ('train', 'val', 'test')
        self.mode = mode
        self.ignore_lb = 255
        self.rootpth = rootpth

        self.imgs = os.listdir(os.path.join(self.rootpth, 'CelebA-HQ-img'))

        #  pre-processing
        self.to_tensor = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ])
        self.trans_train = Compose([
            ColorJitter(
                brightness=0.5,
                contrast=0.5,
                saturation=0.5),
            HorizontalFlip(),
            RandomScale((0.75, 1.0, 1.25, 1.5, 1.75, 2.0)),
            RandomCrop(cropsize)
            ])

    def __getitem__(self, idx):
        impth = self.imgs[idx]
        img = Image.open(osp.join(self.rootpth, 'CelebA-HQ-img', impth))
        img = img.resize((512, 512), Image.BILINEAR)
        label = Image.open(osp.join(self.rootpth, 'mask', impth[:-3]+'png')).convert('P')
        # print(np.unique(np.array(label)))
        if self.mode == 'train':
            im_lb = dict(im=img, lb=label)
            im_lb = self.trans_train(im_lb)
            img, label = im_lb['im'], im_lb['lb']
        img = self.to_tensor(img)
        label = np.array(label).astype(np.int64)[np.newaxis, :]
        return img, label

    def __len__(self):
        return len(self.imgs)


if __name__ == "__main__":
    face_data = '/home/zll/data/CelebAMask-HQ/CelebA-HQ-img'
    face_sep_mask = '/home/zll/data/CelebAMask-HQ/CelebAMask-HQ-mask-anno'
    mask_path = '/home/zll/data/CelebAMask-HQ/mask'
    counter = 0
    total = 0
    for i in range(15):
        # files = os.listdir(osp.join(face_sep_mask, str(i)))

        atts = ['skin', 'l_brow', 'r_brow', 'l_eye', 'r_eye', 'eye_g', 'l_ear', 'r_ear', 'ear_r',
                'nose', 'mouth', 'u_lip', 'l_lip', 'neck', 'neck_l', 'cloth', 'hair', 'hat']

        for j in range(i*2000, (i+1)*2000):

            mask = np.zeros((512, 512))

            for l, att in enumerate(atts, 1):
                total += 1
                file_name = ''.join([str(j).rjust(5, '0'), '_', att, '.png'])
                path = osp.join(face_sep_mask, str(i), file_name)

                if os.path.exists(path):
                    counter += 1
                    sep_mask = np.array(Image.open(path).convert('P'))
                    # print(np.unique(sep_mask))

                    mask[sep_mask == 225] = l
            cv2.imwrite('{}/{}.png'.format(mask_path, j), mask)
            print(j)

    print(counter, total)

















#!/usr/bin/python
# -*- encoding: utf-8 -*-


import torch
import logging

logger = logging.getLogger()

class Optimizer(object):
    def __init__(self,
                model,
                lr0,
                momentum,
                wd,
                warmup_steps,
                warmup_start_lr,
                max_iter,
                power,
                *args, **kwargs):
        self.warmup_steps = warmup_steps
        self.warmup_start_lr = warmup_start_lr
        self.lr0 = lr0
        self.lr = self.lr0
        self.max_iter = float(max_iter)
        self.power = power
        self.it = 0
        wd_params, nowd_params, lr_mul_wd_params, lr_mul_nowd_params = model.get_params()
        param_list = [
                {'params': wd_params},
                {'params': nowd_params, 'weight_decay': 0},
                {'params': lr_mul_wd_params, 'lr_mul': True},
                {'params': lr_mul_nowd_params, 'weight_decay': 0, 'lr_mul': True}]
        self.optim = torch.optim.SGD(
                param_list,
                lr = lr0,
                momentum = momentum,
                weight_decay = wd)
        self.warmup_factor = (self.lr0/self.warmup_start_lr)**(1./self.warmup_steps)


    def get_lr(self):
        if self.it <= self.warmup_steps:
            lr = self.warmup_start_lr*(self.warmup_factor**self.it)
        else:
            factor = (1-(self.it-self.warmup_steps)/(self.max_iter-self.warmup_steps))**self.power
            lr = self.lr0 * factor
        return lr


    def step(self):
        self.lr = self.get_lr()
        for pg in self.optim.param_groups:
            if pg.get('lr_mul', False):
                pg['lr'] = self.lr * 10
            else:
                pg['lr'] = self.lr
        if self.optim.defaults.get('lr_mul', False):
            self.optim.defaults['lr'] = self.lr * 10
        else:
            self.optim.defaults['lr'] = self.lr
        self.it += 1
        self.optim.step()
        if self.it == self.warmup_steps+2:
            logger.info('==> warmup done, start to implement poly lr strategy')

    def zero_grad(self):
        self.optim.zero_grad()




import h5py
import numpy as np
from pyVHR.datasets.dataset import Dataset
from pyVHR.BPM.BPM import BVPsignal


class COHFACE(Dataset):
    """
    Cohface Dataset
    
    .. Cohface dataset structure:
    .. -----------------
    ..    datasetDIR/
    ..    |
    ..    |-- subjDIR_1/
    ..    |   |-- vidDIR1/
    ..    |       |-- videoFile1.avi
    ..    |       |-- ...
    ..    |       |-- videoFileN.avi
    ..    |...
    ..    |   |-- vidDIRM/
    ..    |       |-- videoFile1.avi
    ..    |       |-- ...
    ..    |       |-- videoFileM.avi
    ..    |...
    ..    |-- subjDIR_n/
    ..    |...
    """
    name = 'COHFACE'
    signalGT = 'BVP'         # GT signal type
    numLevels = 2            # depth of the filesystem collecting video and BVP files
    numSubjects = 40         # number of subjects
    video_EXT = 'avi'        # extension of the video files
    frameRate = 20           # vieo frame rate
    VIDEO_SUBSTRING = 'data'  # substring contained in the filename
    SIG_EXT = 'hdf5'         # extension of the BVP files
    SIG_SUBSTRING = 'data'   # substring contained in the filename
    SIG_SampleRate = 256     # sample rate of the BVP files
    skinThresh = [40, 60]     # thresholds for skin detection

    def readSigfile(self, filename):
        """ 
        Load signal from file.
        
        Returns:
            a :py:class:`pyVHR.BPM.BPM.BVPsignal` object that can be used to extract BPM signal from ground truth BVP signal.
        """

        f = h5py.File(filename, 'r')
        data = np.array(f['pulse'])        # load the signal
        data = data.reshape(1, len(data))   # monodimentional signal

        return BVPsignal(data, self.SIG_SampleRate)



import configparser
import ast
from numpy.lib.arraysetops import isin
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import plotly.graph_objects as go
from importlib import import_module, util
from pyVHR.datasets.dataset import datasetFactory
from pyVHR.utils.errors import getErrors, printErrors, displayErrors
from pyVHR.extraction.sig_processing import *
from pyVHR.extraction.sig_extraction_methods import *
from pyVHR.extraction.skin_extraction_methods import *
from pyVHR.BVP.BVP import *
from pyVHR.BPM.BPM import *
from pyVHR.BVP.methods import *
from pyVHR.BVP.filters import *


class MultiMethodSuite():
    """ 
    This class performs tests on a video dataset using multiple rPPG methods.
    The suite uses these methods individually to calculate BVP and estimate BPM. 
    After this all the BVP signals of the methods are combined into a single n-estimator signal, 
    which is used to estimate a last BPM signal. 
    Using multiple methods together allows you to have many estimators, 
    so it can be useful to have a more precise BPM estimate.
    You can customize all the parameters using a configuration file (.cfg); in the
    analysis module you can find an example of cfg file named "multi_method_cfg.cfg".
    """

    def __init__(self, configFilename='default'):
        self.configFilename = configFilename
        if configFilename == 'default':
            self.configFilename = '../pyVHR/analysis/newcfg.cfg'
        self.parse_cfg(self.configFilename)

    def start(self, verb=0):
        """ 
        Runs the tests as specified in the loaded config file.

        Args:
            verb:
               - 0 - not verbose
               - 1 - show the main steps
               
               (use also combinations)
        """

        # -- cfg parser
        parser = configparser.ConfigParser(
            inline_comment_prefixes=('#', ';'))
        parser.optionxform = str
        if not parser.read(self.configFilename):
            raise FileNotFoundError(self.configFilename)

        # -- verbose prints
        if '1' in str(verb):
            self.__verbose('a')

        # -- dataset & cfg params
        if 'path' in self.datasetdict and self.datasetdict['path'] != 'None':
            dataset = datasetFactory(
                self.datasetdict['dataset'], videodataDIR=self.datasetdict['videodataDIR'], BVPdataDIR=self.datasetdict['BVPdataDIR'], path=self.datasetdict['path'])
        else:
            dataset = datasetFactory(
                self.datasetdict['dataset'], videodataDIR=self.datasetdict['videodataDIR'], BVPdataDIR=self.datasetdict['BVPdataDIR'])

        # -- catch data (object)
        res = TestResult()

        # -- SIG processing
        sig_processing = SignalProcessing()
        if bool(self.sigdict['cuda']):
            sig_processing.display_cuda_device()
            sig_processing.choose_cuda_device(int(self.sigdict['cuda_device']))
        # set skin extractor
        target_device = 'GPU' if bool(self.sigdict['cuda']) else 'CPU'
        if self.sigdict['skin_extractor'] == 'convexhull':
            sig_processing.set_skin_extractor(
                SkinExtractionConvexHull(target_device))
        elif self.sigdict['skin_extractor'] == 'faceparsing':
            sig_processing.set_skin_extractor(
                SkinExtractionFaceParsing(target_device))
        # set patches
        if self.sigdict['approach'] == 'patches':
            ldmks_list = ast.literal_eval(
                self.sigdict['landmarks_list'])
            if len(ldmks_list) > 0:
                sig_processing.set_landmarks(ldmks_list)
            if self.sigdict['patches'] == 'squares':
                # set squares patches side dimension
                sig_processing.set_square_patches_side(
                    float(self.sigdict['squares_dim']))
            elif self.sigdict['patches'] == 'rects':
                # set rects patches sides dimensions
                rects_dims = ast.literal_eval(
                    self.sigdict['rects_dims'])
                if len(rects_dims) > 0:
                    sig_processing.set_rect_patches_sides(
                        np.array(rects_dims, dtype=np.float32))
        # set sig-processing and skin-processing params
        SignalProcessingParams.RGB_LOW_TH = np.int32(
            self.sigdict['sig_color_low_threshold'])
        SignalProcessingParams.RGB_HIGH_TH = np.int32(
            self.sigdict['sig_color_high_threshold'])
        SkinProcessingParams.RGB_LOW_TH = np.int32(
            self.sigdict['skin_color_low_threshold'])
        SkinProcessingParams.RGB_HIGH_TH = np.int32(
            self.sigdict['skin_color_high_threshold'])

        # load all the videos
        if self.videoIdx == []:
            self.videoIdx = [int(v)
                             for v in range(len(dataset.videoFilenames))]

        # -- loop on videos
        for v in self.videoIdx:

            # -- verbose prints
            if '1' in str(verb):
                print("\n## videoID: %d" % (v))

            # -- ground-truth signal
            try:
                fname = dataset.getSigFilename(v)
                sigGT = dataset.readSigfile(fname)
            except:
                continue
            winSizeGT = int(self.sigdict['winSize'])
            bpmGT, timesGT = sigGT.getBPM(winSizeGT)

            # -- video file name
            videoFileName = dataset.getVideoFilename(v)
            print(videoFileName)
            fps = get_fps(videoFileName)

            sig_processing.set_total_frames(
                int(self.sigdict['tot_sec'])*fps)

            sig = []
            if str(self.sigdict['approach']) == 'hol':
                # SIG extraction with holistic
                sig = sig_processing.extract_holistic(videoFileName)
            elif str(self.sigdict['approach']) == 'patches':
                # SIG extraction with patches
                sig = sig_processing.extract_patches(
                    videoFileName, str(self.sigdict['patches']), str(self.sigdict['type']))

            # -- sig windowing
            windowed_sig, timesES = sig_windowing(
                sig, int(self.sigdict['winSize']), 1, fps)

            # -- loop on methods
            bvps_collection = []
            methods_collection = []
            for m in self.methods:
                if '1' in str(verb):
                    print("## Computing method: %s" % (str(m)))
                # -- PRE FILTERING
                filtered_windowed_sig = windowed_sig

                # -- color threshold - applied only with patches
                if str(self.sigdict['approach']) == 'patches':
                    filtered_windowed_sig = apply_filter(
                        windowed_sig,
                        rgb_filter_th,
                        params={'RGB_LOW_TH':  np.int32(self.bvpdict['color_low_threshold']),
                                'RGB_HIGH_TH': np.int32(self.bvpdict['color_high_threshold'])})

                # -- custom filters
                prefilter_list = ast.literal_eval(
                    self.methodsdict[m]['pre_filtering'])
                if len(prefilter_list) > 0:
                    for f in prefilter_list:
                        if '1' in str(verb):
                            print("  pre-filter: %s" % f)
                        fdict = dict(parser[f].items())
                        if fdict['path'] != 'None':
                            # custom path
                            spec = util.spec_from_file_location(
                                fdict['name'], fdict['path'])
                            mod = util.module_from_spec(spec)
                            spec.loader.exec_module(mod)
                            method_to_call = getattr(
                                mod, fdict['name'])
                        else:
                            # package path
                            module = import_module(
                                'pyVHR.BVP.filters')
                            method_to_call = getattr(
                                module, fdict['name'])
                        filtered_windowed_sig = apply_filter(
                            filtered_windowed_sig, method_to_call, fps=fps, params=ast.literal_eval(fdict['params']))

                # -- BVP extraction
                if self.methodsdict[m]['path'] != 'None':
                    # custom path
                    spec = util.spec_from_file_location(
                        self.methodsdict[m]['name'], self.methodsdict[m]['path'])
                    mod = util.module_from_spec(spec)
                    spec.loader.exec_module(mod)
                    method_to_call = getattr(mod, self.methodsdict[m]['name'])
                else:
                    # package path
                    module = import_module(
                        'pyVHR.BVP.methods')
                    method_to_call = getattr(
                        module, self.methodsdict[m]['name'])
                bvps = RGB_sig_to_BVP(filtered_windowed_sig, fps,
                                      device_type=self.methodsdict[m]['device_type'], method=method_to_call, params=ast.literal_eval(self.methodsdict[m]['params']))

                # POST FILTERING
                postfilter_list = ast.literal_eval(
                    self.methodsdict[m]['post_filtering'])
                if len(postfilter_list) > 0:
                    for f in postfilter_list:
                        if '1' in str(verb):
                            print("  post-filter: %s" % f)
                        fdict = dict(parser[f].items())
                        if fdict['path'] != 'None':
                            # custom path
                            spec = util.spec_from_file_location(
                                fdict['name'], fdict['path'])
                            mod = util.module_from_spec(spec)
                            spec.loader.exec_module(mod)
                            method_to_call = getattr(
                                mod, fdict['name'])
                        else:
                            # package path
                            module = import_module(
                                'pyVHR.BVP.filters')
                            method_to_call = getattr(
                                module, fdict['name'])
                        bvps = apply_filter(
                            bvps, method_to_call, fps=fps, params=ast.literal_eval(fdict['params']))

                bvps_collection.append(bvps)
                methods_collection.append(m)

            # -- BPM extraction
            # first each method, then multi-method
            multi_method_coll = [e for e in bvps_collection]
            methods_collection.append("Multi-Method")
            for i in range(len(bvps_collection)+1):
                if i == len(bvps_collection): # last element is the multi-method
                    bvp_element = concatenate_BVPs(multi_method_coll)
                    if bvp_element == 0:
                        print("[ERROR] Multi method Concatenation can't be executed because BVP signals have different shapes!")
                        continue
                else: # use single methods
                    bvp_element = bvps_collection[i]

                bpmES = None
                if str(self.bpmdict['type']) == 'welch':
                    bpmES = BVP_to_BPM(bvp_element, fps, minHz=float(
                        self.bpmdict['minHz']), maxHz=float(self.bpmdict['maxHz']))
                elif str(self.bpmdict['type']) == 'psd_clustering':
                    bpmES = BVP_to_BPM_PSD_clustering(bvp_element, fps, minHz=float(
                        self.bpmdict['minHz']), maxHz=float(self.bpmdict['maxHz']))
                if bpmES is None:
                    print("[ERROR] BPM extraction error; check cfg params!")
                    continue
                # median BPM from multiple estimators BPM
                # this doesn't affect holistic approach
                median_bpmES = multi_est_BPM_median(bpmES)

                # -- error metrics
                RMSE, MAE, MAX, PCC = getErrors(median_bpmES, bpmGT, timesES, timesGT)

                # -- save results
                method_name = methods_collection[i]
                res.newDataSerie()
                res.addData('dataset', str(self.datasetdict['dataset']))
                res.addData('method', str(method_name))
                res.addData('videoIdx', v)
                res.addData('RMSE', RMSE)
                res.addData('MAE', MAE)
                res.addData('MAX', MAX)
                res.addData('PCC', PCC)
                res.addData('bpmGT', bpmGT)
                res.addData('bpmES', median_bpmES)
                res.addData('timeGT', timesGT)
                res.addData('timeES', timesES)
                res.addData('videoFilename', videoFileName)
                res.addDataSerie()

                if '1' in str(verb):
                    print("## Results for method: %s" % (str(method_name)))
                    printErrors(RMSE, MAE, MAX, PCC)

        return res

    def parse_cfg(self, configFilename):
        """ parses the given configuration file for loading the test's parameters.
        
        Args:
            configFilename: configuation file (.cfg) name of path .

        """

        self.parser = configparser.ConfigParser(
            inline_comment_prefixes=('#', ';'))
        self.parser.optionxform = str
        if not self.parser.read(configFilename):
            raise FileNotFoundError(configFilename)

        # load paramas
        self.datasetdict = dict(self.parser['DATASET'].items())
        self.sigdict = dict(self.parser['SIG'].items())
        self.bvpdict = dict(self.parser['BVP'].items())
        self.bpmdict = dict(self.parser['BPM'].items())

        # video idx list extraction
        if isinstance(ast.literal_eval(self.datasetdict['videoIdx']), list):
            self.videoIdx = [int(v) for v in ast.literal_eval(
                self.datasetdict['videoIdx'])]

        # load parameters for each methods
        self.methodsdict = {}
        self.methods = ast.literal_eval(self.bvpdict['methods'])
        for x in self.methods:
            self.methodsdict[x] = dict(self.parser[x].items())

    def __merge(self, dict1, dict2):
        for key in dict2:
            if key not in dict1:
                dict1[key] = dict2[key]

    def __verbose(self, verb):
        if verb == 'a':
            print("** Run the test with the following config:")
            print("      dataset: " + self.datasetdict['dataset'].upper())
            print("      methods: " + str(self.methods))


class TestResult():
    """ 
    This class is used by :py:class:`pyVHR.analysis.multi_method_suite.MultiMethodSuite` to manage the results
    of a test for a given video dataset on multiple rPPG methods
    """

    def __init__(self, filename=None):

        if filename == None:
            self.dataFrame = pd.DataFrame()
        else:
            self.dataFrame = pd.read_hdf(filename)
        self.dict = None

    def addDataSerie(self):
        # -- store serie
        if self.dict != None:
            self.dataFrame = self.dataFrame.append(
                self.dict, ignore_index=True)

    def newDataSerie(self):
        # -- new dict
        D = {}
        D['method'] = ''
        D['dataset'] = ''
        D['videoIdx'] = ''        # video filename
        D['sigFilename'] = ''     # GT signal filename
        D['videoFilename'] = ''   # GT signal filename
        D['RMSE'] = ''
        D['MAE'] = ''
        D['PCC'] = ''
        D['MAX'] = ''
        D['bpmGT'] = ''          # GT bpm
        D['bpmES'] = ''
        D['timeGT'] = ''            # GT bpm
        D['timeES'] = ''
        self.dict = D

    def addData(self, key, value):
        self.dict[key] = value

    def saveResults(self, outFilename=None):
        """
        Save the test results in a HDF5 library that can be opened using pandas.
        You can analyze the results using :py:class:`pyVHR.analysis.stats.StatAnalysis`
        """
        if outFilename == None:
            outFilename = "testResults.h5"
        else:
            self.outFilename = outFilename

        # -- save data
        self.dataFrame.to_hdf(outFilename, key='df', mode='w')



from threading import Thread
import time
from time import sleep
from pyVHR.realtime.params import Params
from pyVHR.realtime.video_capture import VideoCapture
import cv2
import numpy as np
import plotly.graph_objects as go
from importlib import import_module, util
from pyVHR.datasets.dataset import datasetFactory
from pyVHR.utils.errors import getErrors, printErrors, displayErrors
from pyVHR.extraction.sig_processing import *
from pyVHR.extraction.sig_extraction_methods import *
from pyVHR.extraction.skin_extraction_methods import *
from pyVHR.BVP.BVP import *
from pyVHR.BPM.BPM import *
from pyVHR.BVP.methods import *
from pyVHR.BVP.filters import *
import PySimpleGUI as sg
import numpy as np
import plotly.graph_objects as go
import queue
import statistics


class SharedData:
    def __init__(self):
        self.q_bpm = queue.Queue()
        self.q_video_image = queue.Queue()
        self.q_skin_image = queue.Queue()
        self.q_patches_image = queue.Queue()
        self.q_stop = queue.Queue()
        self.q_stop_cap = queue.Queue()
        self.q_frames = queue.Queue()


def VHRroutine(sharedData):
    sig_ext_met = None
    ldmks_regions = None
    # Holistic settings #
    if Params.approach == 'holistic':
        sig_ext_met = holistic_mean
    # Patches settings #
    elif Params.approach == 'patches':
        # extraction method
        if Params.type == "mean" and Params.patches == "squares":
            sig_ext_met = landmarks_mean
        elif Params.type == "mean" and Params.patches == "rects":
            sig_ext_met = landmarks_mean_custom_rect
        elif Params.type == "median" and Params.patches == "squares":
            sig_ext_met = landmarks_median
        elif Params.type == "median" and Params.patches == "rects":
            sig_ext_met = landmarks_median_custom_rect
        # patches dims
        if Params.patches == "squares":
            ldmks_regions = np.float32(Params.squares_dim)
        elif Params.patches == "rects":
            ldmks_regions = np.float32(Params.rects_dims)

    SignalProcessingParams.RGB_LOW_TH = np.int32(
        Params.sig_color_low_threshold)
    SignalProcessingParams.RGB_HIGH_TH = np.int32(
        Params.sig_color_high_threshold)
    SkinProcessingParams.RGB_LOW_TH = np.int32(
        Params.skin_color_low_threshold)
    SkinProcessingParams.RGB_HIGH_TH = np.int32(
        Params.skin_color_high_threshold)

    color = np.array([Params.font_color[0],
                      Params.font_color[1], Params.font_color[2]], dtype=np.uint8)

    skin_ex = None
    if Params.skin_extractor == 'convexhull':
        skin_ex = SkinExtractionConvexHull()
    elif Params.skin_extractor == 'faceparsing':
        skin_ex = SkinExtractionFaceParsing()

    mp_drawing = mp.solutions.drawing_utils
    mp_face_mesh = mp.solutions.face_mesh
    PRESENCE_THRESHOLD = 0.5
    VISIBILITY_THRESHOLD = 0.5

    if Params.fps_fixed is not None:
        fps = Params.fps_fixed
    else:
        fps = get_fps(Params.videoFileName)
    tot_frames = int(Params.tot_sec*fps)

    sig = []
    processed_frames_count = 0
    sig_buff_dim = int(fps * Params.winSize)
    sig_stride = int(fps * Params.stride)
    sig_buff_counter = sig_stride

    BPM_obj = None

    timeCount = []

    cap = VideoCapture(Params.videoFileName, sharedData, fps=fps,
                       sleep=Params.fake_delay, resize=Params.resize)

    send_images_count = 0
    send_images_stride = 3

    with mp_face_mesh.FaceMesh(
            max_num_faces=1,
            min_detection_confidence=0.5,
            min_tracking_confidence=0.5) as face_mesh:
        while True:
            start_time = time.perf_counter()*1000
            frame = None
            if not sharedData.q_frames.empty():  # read frames from shared data
                frame = sharedData.q_frames.get()
                if type(frame) == int:  # cap stopped
                    break
            if not sharedData.q_stop.empty():  # GUI stopped
                sharedData.q_stop.get()
                break
            if frame is None:
                continue
            # convert the BGR image to RGB.
            image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            processed_frames_count += 1
            width = image.shape[1]
            height = image.shape[0]
            # [landmarks, info], with info->x_center ,y_center, r, g, b
            ldmks = np.zeros((468, 5), dtype=np.float32)
            ldmks[:, 0] = -1.0
            ldmks[:, 1] = -1.0
            magic_ldmks = []
            ### face landmarks ###
            results = face_mesh.process(image)
            if results.multi_face_landmarks:
                face_landmarks = results.multi_face_landmarks[0]
                landmarks = [l for l in face_landmarks.landmark]
                for idx in range(len(landmarks)):
                    landmark = landmarks[idx]
                    if not ((landmark.HasField('visibility') and landmark.visibility < VISIBILITY_THRESHOLD)
                            or (landmark.HasField('presence') and landmark.presence < PRESENCE_THRESHOLD)):
                        coords = mp_drawing._normalized_to_pixel_coordinates(
                            landmark.x, landmark.y, width, height)
                        if coords:
                            ldmks[idx, 0] = coords[1]
                            ldmks[idx, 1] = coords[0]
                ### skin extraction ###
                cropped_skin_im, full_skin_im = skin_ex.extract_skin(
                    image, ldmks)
            else:
                cropped_skin_im = np.zeros_like(image)
                full_skin_im = np.zeros_like(image)
            ### SIG ###
            if Params.approach == 'patches':
                magic_ldmks = np.array(
                    ldmks[Params.landmarks_list], dtype=np.float32)
                temp = sig_ext_met(magic_ldmks, full_skin_im, ldmks_regions,
                                   np.int32(SignalProcessingParams.RGB_LOW_TH), np.int32(SignalProcessingParams.RGB_HIGH_TH))
                temp = temp[:, 2:]  # keep only rgb mean
            elif Params.approach == 'holistic':
                temp = sig_ext_met(cropped_skin_im, np.int32(
                    SignalProcessingParams.RGB_LOW_TH), np.int32(SignalProcessingParams.RGB_HIGH_TH))
            sig.append(temp)

            # visualize original, patches and skin
            if send_images_count == send_images_stride:
                send_images_count = 0
                sharedData.q_video_image.put(image)
                if Params.visualize_skin == True:
                    sharedData.q_skin_image.put(full_skin_im)
                if Params.approach == 'patches' and Params.visualize_landmarks == True:
                    annotated_image = full_skin_im.copy()
                    for idx in Params.landmarks_list:
                        cv2.circle(
                            annotated_image, (int(ldmks[idx, 1]), int(ldmks[idx, 0])), radius=0, color=Params.font_color, thickness=-1)
                        if Params.visualize_landmarks_number == True:
                            cv2.putText(annotated_image, str(idx),
                                        (int(ldmks[idx, 1]), int(ldmks[idx, 0])), cv2.FONT_HERSHEY_SIMPLEX, Params.font_size,  Params.font_color,  1)
                    if Params.visualize_patches == True:
                        if Params.patches == "squares":
                            sides = [Params.squares_dim, ] * len(magic_ldmks)
                            sides = np.array(sides)
                            annotated_image = draw_rects(
                                annotated_image, np.array(magic_ldmks[:, 1]), np.array(magic_ldmks[:, 0]), sides, sides, color)
                        elif Params.patches == "rects":
                            rects_dims = np.array(Params.rects_dims)
                            annotated_image = draw_rects(
                                annotated_image, np.array(magic_ldmks[:, 1]),
                                np.array(magic_ldmks[:, 0]), rects_dims[:, 0], rects_dims[:, 1], color)
                    # visualize patches
                    sharedData.q_patches_image.put(annotated_image)
            else:
                send_images_count += 1

            if processed_frames_count > sig_buff_dim:
                sig = sig[1:]
                if sig_buff_counter == 0:
                    sig_buff_counter = sig_stride
                    copy_sig = np.array(sig, dtype=np.float32)
                    copy_sig = np.swapaxes(copy_sig, 0, 1)
                    copy_sig = np.swapaxes(copy_sig, 1, 2)

                    ### Pre_filtering ###
                    if Params.approach == 'patches':
                        copy_sig = rgb_filter_th(copy_sig, **{'RGB_LOW_TH':  np.int32(Params.color_low_threshold),
                                                              'RGB_HIGH_TH': np.int32(Params.color_high_threshold)})
                    for filt in Params.pre_filter:
                        if filt != {}:
                            if 'fps' in filt['params'] and filt['params']['fps'] == 'adaptive' and fps is not None:
                                filt['params']['fps'] = float(fps)
                            if filt['params'] == {}:
                                copy_sig = filt['filter_func'](
                                    copy_sig)
                            else:
                                copy_sig = filt['filter_func'](
                                    copy_sig, **filt['params'])

                    ### BVP ###
                    bvp = np.zeros((0, 1), dtype=np.float32)
                    if Params.method['device_type'] == 'cpu':
                        bvp = signals_to_bvps_cpu(
                            copy_sig, Params.method['method_func'], Params.method['params'])
                    elif Params.method['device_type'] == 'torch':
                        bvp = signals_to_bvps_torch(
                            copy_sig, Params.method['method_func'], Params.method['params'])

                    ### Post_filtering ###
                    for filt in Params.pre_filter:
                        if filt != {}:
                            bvp = np.expand_dims(bvp, axis=1)
                            if 'fps' in filt['params'] and filt['params']['fps'] == 'adaptive' and fps is not None:
                                filt['params']['fps'] = float(fps)
                            if filt['params'] == {}:
                                bvp = filt['filter_func'](bvp)
                            else:
                                bvp = filt['filter_func'](
                                    bvp, **filt['params'])
                            bvp = np.squeeze(bvp, axis=1)

                    ### BPM ###
                    if BPM_obj == None:
                        BPM_obj = BPM(bvp, fps, minHz=Params.minHz,
                                        maxHz=Params.maxHz)
                    else:
                        BPM_obj.data = bvp
                    if Params.BPM_extraction_type == "welch":
                        bpm = BPM_obj.BVP_to_BPM()
                    elif Params.BPM_extraction_type == "psd_clustering":
                        bpm = BPM_obj.BVP_to_BPM_PSD_clustering()
                    if Params.approach == 'patches':  # Median of multi BPMs
                        if len(bpm.shape) > 0 and bpm.shape[0] == 0:
                            bpm = np.float32(0.0)
                        else:
                            bpm = np.float32(np.median(bpm))
                    sharedData.q_bpm.put(bpm)
                else:
                    sig_buff_counter -= 1

            end_time = time.perf_counter()*1000
            timeCount.append(end_time-start_time)
            if len(timeCount) > 100:
                timeCount = timeCount[1:]

            ### loop break ###
            if tot_frames is not None and tot_frames > 0 and processed_frames_count >= tot_frames:
                break
    if len(timeCount) > 2:
        print("Times in milliseconds of the computation of a frame:")
        print("mean:   ",statistics.mean(timeCount))
        print("median: ",statistics.median(timeCount))
        print("max:    ",max(timeCount))
        print("min:    ",min(timeCount))
    return


"""
if __name__ == "__main__":
    sd = SharedData()
    Params.videoFileName = "/home/frea/Documents/VHR/LGI_PPGI/lgi_alex/alex_resting/cv_camera_sensor_stream_handler.avi"
    Params.tot_sec = 0
    t = Thread(target=VHRroutine, args=(sd,))
    t.start()
    t.join()
"""



import cv2
import threading
import time
from pyVHR.extraction.utils import *


class VideoCapture:
    # bufferless VideoCapture
    def __init__(self, name, sharedData, fps=None, sleep=False, resize=True):
        self.cap = cv2.VideoCapture(name)
        self.sleep = sleep
        self.resize = resize
        self.fps = None
        if fps is not None:
            self.cap.set(cv2.CAP_PROP_FPS, fps)
            self.fps = fps
        self.sd = sharedData
        self.t = threading.Thread(target=self._reader)
        self.t.daemon = False  # when process ends all deamon will be killed
        self.t.start()

    # read frames as soon as they are available
    def _reader(self):
        while True:
            if not self.sd.q_stop_cap.empty():  # GUI stopped
                self.sd.q_stop_cap.get()
                self.sd.q_stop.put(0)  # stop VHR model
                self.sd.q_frames.put(0)
                break
            ret, frame = self.cap.read()
            if not ret:  # IO error
                self.sd.q_stop.put(0)  # stop VHR model
                self.sd.q_frames.put(0)  # end extraction -> send number
                break
            if self.resize:
                h, w = frame.shape[0], frame.shape[1]
                if h != 480 or w != 640:
                    frame = cv2.resize(frame, (640, int(640*h/w)),
                                       interpolation=cv2.INTER_NEAREST)
            self.sd.q_frames.put(frame)
            if self.sleep and self.fps is not None:
                time.sleep(self.fps / 1000.0)
        self.cap.release()


"""
if __name__ == "__main__":
    import cv2
    from pyVHR.realtime.VHRroutine import SharedData
    sd = SharedData()
    cap = VideoCapture(
        "/home/frea/Documents/VHR/LGI_PPGI/lgi_alex/alex_resting/cv_camera_sensor_stream_handler.avi", fps=25, sharedData=sd)

    # Check if the webcam is opened correctly
    for i in range(10):
        frame = sd.q_frames.get()
        #cv2.imshow('Input', frame)
        print(frame.shape)
    sd.q_stop_cap.put(0)
    time.sleep(2)
"""



#!/usr/bin/python
# -*- encoding: utf-8 -*-


import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np


class OhemCELoss(nn.Module):
    def __init__(self, thresh, n_min, ignore_lb=255, *args, **kwargs):
        super(OhemCELoss, self).__init__()
        self.thresh = -torch.log(torch.tensor(thresh, dtype=torch.float)).cuda()
        self.n_min = n_min
        self.ignore_lb = ignore_lb
        self.criteria = nn.CrossEntropyLoss(ignore_index=ignore_lb, reduction='none')

    def forward(self, logits, labels):
        N, C, H, W = logits.size()
        loss = self.criteria(logits, labels).view(-1)
        loss, _ = torch.sort(loss, descending=True)
        if loss[self.n_min] > self.thresh:
            loss = loss[loss>self.thresh]
        else:
            loss = loss[:self.n_min]
        return torch.mean(loss)


class SoftmaxFocalLoss(nn.Module):
    def __init__(self, gamma, ignore_lb=255, *args, **kwargs):
        super(SoftmaxFocalLoss, self).__init__()
        self.gamma = gamma
        self.nll = nn.NLLLoss(ignore_index=ignore_lb)

    def forward(self, logits, labels):
        scores = F.softmax(logits, dim=1)
        factor = torch.pow(1.-scores, self.gamma)
        log_score = F.log_softmax(logits, dim=1)
        log_score = factor * log_score
        loss = self.nll(log_score, labels)
        return loss


if __name__ == '__main__':
    torch.manual_seed(15)
    criteria1 = OhemCELoss(thresh=0.7, n_min=16*20*20//16).cuda()
    criteria2 = OhemCELoss(thresh=0.7, n_min=16*20*20//16).cuda()
    net1 = nn.Sequential(
        nn.Conv2d(3, 19, kernel_size=3, stride=2, padding=1),
    )
    net1.cuda()
    net1.train()
    net2 = nn.Sequential(
        nn.Conv2d(3, 19, kernel_size=3, stride=2, padding=1),
    )
    net2.cuda()
    net2.train()

    with torch.no_grad():
        inten = torch.randn(16, 3, 20, 20).cuda()
        lbs = torch.randint(0, 19, [16, 20, 20]).cuda()
        lbs[1, :, :] = 255

    logits1 = net1(inten)
    logits1 = F.interpolate(logits1, inten.size()[2:], mode='bilinear')
    logits2 = net2(inten)
    logits2 = F.interpolate(logits2, inten.size()[2:], mode='bilinear')

    loss1 = criteria1(logits1, lbs)
    loss2 = criteria2(logits2, lbs)
    loss = loss1 + loss2
    print(loss.detach().cpu())
    loss.backward()



from numba import prange, njit
import numpy as np
import cv2

def get_magic_landmarks():
    """ returns high_priority and mid_priority list of landmarks identification number """
    return [*MagicLandmarks.forehead_center, *MagicLandmarks.cheek_left_bottom, *MagicLandmarks.cheek_right_bottom], [*MagicLandmarks.forehoead_right, *MagicLandmarks.forehead_left, *MagicLandmarks.cheek_left_top, *MagicLandmarks.cheek_right_top]


class MagicLandmarks():
    """
    This class contains usefull lists of landmarks identification numbers.
    """
    high_prio_forehead = [10, 67, 69, 104, 108, 109, 151, 299, 337, 338]
    high_prio_nose = [3, 4, 5, 6, 45, 51, 115, 122, 131, 134, 142, 174, 195, 196, 197, 198,
                      209, 217, 220, 236, 248, 275, 277, 281, 360, 363, 399, 419, 420, 429, 437, 440, 456]
    high_prio_left_cheek = [36, 47, 50, 100, 101, 116, 117,
                            118, 119, 123, 126, 147, 187, 203, 205, 206, 207, 216]
    high_prio_right_cheek = [266, 280, 329, 330, 346, 347,
                             347, 348, 355, 371, 411, 423, 425, 426, 427, 436]

    mid_prio_forehead = [8, 9, 21, 68, 103, 251,
                         284, 297, 298, 301, 332, 333, 372, 383]
    mid_prio_nose = [1, 44, 49, 114, 120, 121, 128, 168, 188, 351, 358, 412]
    mid_prio_left_cheek = [34, 111, 137, 156, 177, 192, 213, 227, 234]
    mid_prio_right_cheek = [340, 345, 352, 361, 454]
    mid_prio_chin = [135, 138, 169, 170, 199, 208, 210, 211,
                     214, 262, 288, 416, 428, 430, 431, 432, 433, 434]
    mid_prio_mouth = [92, 164, 165, 167, 186, 212, 322, 391, 393, 410]
    # more specific areas
    forehead_left = [21, 71, 68, 54, 103, 104, 63, 70,
                     53, 52, 65, 107, 66, 108, 69, 67, 109, 105]
    forehead_center = [10, 151, 9, 8, 107, 336, 285, 55, 8]
    forehoead_right = [338, 337, 336, 296, 285, 295, 282,
                       334, 293, 301, 251, 298, 333, 299, 297, 332, 284]
    eye_right = [283, 300, 368, 353, 264, 372, 454, 340, 448,
                 450, 452, 464, 417, 441, 444, 282, 276, 446, 368]
    eye_left = [127, 234, 34, 139, 70, 53, 124,
                35, 111, 228, 230, 121, 244, 189, 222, 143]
    nose = [193, 417, 168, 188, 6, 412, 197, 174, 399, 456,
            195, 236, 131, 51, 281, 360, 440, 4, 220, 219, 305]
    mounth_up = [186, 92, 167, 393, 322, 410, 287, 39, 269, 61, 164]
    mounth_down = [43, 106, 83, 18, 406, 335, 273, 424, 313, 194, 204]
    chin = [204, 170, 140, 194, 201, 171, 175,
            200, 418, 396, 369, 421, 431, 379, 424]
    cheek_left_bottom = [215, 138, 135, 210, 212, 57, 216, 207, 192]
    cheek_right_bottom = [435, 427, 416, 364,
                          394, 422, 287, 410, 434, 436]
    cheek_left_top = [116, 111, 117, 118, 119, 100, 47, 126, 101, 123,
                      137, 177, 50, 36, 209, 129, 205, 147, 177, 215, 187, 207, 206, 203]
    cheek_right_top = [349, 348, 347, 346, 345, 447, 323,
                       280, 352, 330, 371, 358, 423, 426, 425, 427, 411, 376]
    # dense zones used for convex hull masks
    left_eye = [157,144, 145, 22, 23, 25, 154, 31, 160, 33, 46, 52, 53, 55, 56, 189, 190, 63, 65, 66, 70, 221, 222, 223, 225, 226, 228, 229, 230, 231, 232, 105, 233, 107, 243, 124]
    right_eye = [384, 385, 386, 259, 388, 261, 265, 398, 276, 282, 283, 285, 413, 293, 296, 300, 441, 442, 445, 446, 449, 451, 334, 463, 336, 464, 467, 339, 341, 342, 353, 381, 373, 249, 253, 255]
    mounth = [391, 393, 11, 269, 270, 271, 287, 164, 165, 37, 167, 40, 43, 181, 313, 314, 186, 57, 315, 61, 321, 73, 76, 335, 83, 85, 90, 106]
    # equispaced facial points - mouth and eyes are excluded.
    equispaced_facial_points = [2, 3, 4, 5, 6, 8, 9, 10, 18, 21, 32, 35, 36, 43, 46, 47, 48, 50, 54, \
             58, 67, 68, 69, 71, 92, 93, 101, 103, 104, 108, 109, 116, 117, \
             118, 123, 132, 134, 135, 138, 139, 142, 148, 149, 150, 151, 152, 182, 187, 188, 193, 197, 201, 205, 206, 207, \
             210, 211, 212, 216, 234, 248, 251, 262, 265, 266, 273, 277, 278, 280, \
             284, 288, 297, 299, 322, 323, 330, 332, 333, 337, 338, 345, \
             346, 361, 363, 364, 367, 368, 371, 377, 379, 411, 412, 417, 421, 425, 426, 427, 430, 432, 436]

@njit(parallel=True)
def draw_rects(image, xcenters, ycenters, xsides, ysides, color):
    """
    This method is used to draw N rectangles on a image.
    """
    for idx in prange(len(xcenters)):
        leftx = int(xcenters[idx] - xsides[idx]/2)
        rightx = int(xcenters[idx] + xsides[idx]/2)
        topy = int(ycenters[idx] - ysides[idx]/2)
        bottomy = int(ycenters[idx] + ysides[idx]/2)
        for x in prange(leftx, rightx):
            if topy >= 0 and x >= 0 and x < image.shape[1]:
                image[topy, x, 0] = color[0]
                image[topy, x, 1] = color[1]
                image[topy, x, 2] = color[2]
            if bottomy < image.shape[0] and x >= 0 and x < image.shape[1]:
                image[bottomy, x, 0] = color[0]
                image[bottomy, x, 1] = color[1]
                image[bottomy, x, 2] = color[2]
        for y in prange(topy, bottomy):
            if leftx >= 0 and y >= 0 and y < image.shape[0]:
                image[y, leftx, 0] = color[0]
                image[y, leftx, 1] = color[1]
                image[y, leftx, 2] = color[2]
            if rightx < image.shape[1] and y >= 0 and y < image.shape[0]:
                image[y, rightx, 0] = color[0]
                image[y, rightx, 1] = color[1]
                image[y, rightx, 2] = color[2]
    return image

def sig_windowing(sig, wsize, stride, fps):
    """
    This method is used to divide a RGB signal into overlapping windows.

    Args:
        sig (float32 ndarray): ndarray with shape [num_frames, num_estimators, rgb_channels].
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        windowed signal as a list of length num_windows of float32 ndarray with shape [num_estimators, rgb_channels, window_frames],
        and a 1D ndarray of times in seconds,where each one is the center of a window.
    """
    N = sig.shape[0]
    block_idx, timesES = sliding_straded_win_offline(N, wsize, stride, fps)
    block_signals = []
    for e in block_idx:
        st_frame = int(e[0])
        end_frame = int(e[-1])
        wind_signal = np.copy(sig[st_frame: end_frame+1])
        wind_signal = np.swapaxes(wind_signal, 0, 1)
        wind_signal = np.swapaxes(wind_signal, 1, 2)
        block_signals.append(wind_signal)
    return block_signals, timesES

def raw_windowing(raw_signal, wsize, stride, fps):
    """
    This method is used to divide a Raw signal into overlapping windows.

    Args:
        sig (float32 ndarray): ndarray of images with shape [num_frames, rows, columns, rgb_channels].
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        windowed signal as a list of length num_windows of float32 ndarray with shape [num_frames, rows, columns, rgb_channels],
        and a 1D ndarray of times in seconds,where each one is the center of a window.
    """
    N = raw_signal.shape[0]
    block_idx, timesES = sliding_straded_win_offline(N, wsize, stride, fps)
    block_signals = []
    for e in block_idx:
        st_frame = int(e[0])
        end_frame = int(e[-1])
        wind_signal = np.copy(raw_signal[st_frame: end_frame+1])
        block_signals.append(wind_signal)
    return block_signals, timesES

def sliding_straded_win_offline(N, wsize, stride, fps):
    """
    This method is used to compute all the info for creating an overlapping windows signal.

    Args:
        N (int): length of the signal.
        wsize (float): window size in seconds.
        stride (float): stride between overlapping windows in seconds.
        fps (float): frames per seconds.

    Returns:
        List of ranges, each one contains the indices of a window, and a 1D ndarray of times in seconds, where each one is the center of a window.
    """
    wsize_fr = wsize*fps
    stride_fr = stride*fps
    idx = []
    timesES = []
    num_win = int((N-wsize_fr)/stride_fr)+1
    s = 0
    for i in range(num_win):
        idx.append(np.arange(s, s+wsize_fr))
        s += stride_fr
        timesES.append(wsize/2+stride*i)
    return idx, np.array(timesES, dtype=np.float32)

def get_fps(videoFileName):
    """
    This method returns the fps of a video file name or path.
    """
    vidcap = cv2.VideoCapture(videoFileName)
    fps = vidcap.get(cv2.CAP_PROP_FPS)
    vidcap.release()
    return fps

def extract_frames_yield(videoFileName):
    """
    This method yield the frames of a video file name or path.
    """
    vidcap = cv2.VideoCapture(videoFileName)
    success, image = vidcap.read()
    while success:
        yield image
        success, image = vidcap.read()
    vidcap.release()


import numpy as np
import pybdf
from biosppy.signals import ecg
from pyVHR.datasets.dataset import Dataset
from pyVHR.utils.ecg import ECGsignal

"""
In order to use this module you need a Fortran Compiler.

For Linux you can use:
    sudo apt-get install gfortran

"""

class MAHNOB(Dataset):
    """
    Mahnob Dataset

    .. Mahnob dataset structure:
    .. -----------------
    ..     datasetDIR/
    ..     |
    ..     ||-- vidDIR1/
    ..     |   |-- videoFile.avi
    ..     |   |-- physioFile.bdf
    ..     |...
    ..     |...
    """
    name = 'MAHNOB'
    signalGT = 'ECG'     # GT signal type
    numLevels = 2        # depth of the filesystem collecting video and BVP files
    numSubjects = 40     # number of subjects
    video_EXT = 'avi'    # extension of the video files
    frameRate = 20       # video frame rate
    VIDEO_SUBSTRING = 'Section'  # substring contained in the filename
    SIG_EXT = 'bdf'     # extension of the ECG files
    SIG_SUBSTRING = 'emotion' # substring contained in the filename
    SIG_SampleRate = 256 # sample rate of the ECG files

    def readSigfile(self, filename):
        """ Load ECG signal.

            Returns:
                a pyVHR.utils.ecg.ECGsignal object that can be used to extract ground truth BPM signal.
            
        """
        bdfRec = pybdf.bdfRecording(filename)
        rec = bdfRec.getData(channels=[33])
        self.SIG_SampleRate = bdfRec.sampRate[33]
        data = np.array(rec['data'][0])[rec['eventTable']['idx'][2]:]
        return ECGsignal(data, self.SIG_SampleRate)
        



from abc import ABCMeta, abstractmethod
import os
from importlib import import_module, util

# def datasetFactory(datasetName, *args, **kwargs):


def datasetFactory(datasetName, videodataDIR, BVPdataDIR, path=None):
    """
    This method is used for creating a new istance of a dataset Class (that
    innherit :py:class:`pyVHR.datasets.dataset.Dataset` ).

    Args:
        datasetName (str): name of the dataset Class.
        videodataDIR (str): path of the video data directory.
        BVPdataDIR (str): path of the ground truth BVP data directory.
    """
    try:
        if path == None:
            moduleName = datasetName.lower()
            className = datasetName.upper()
            datasetModule = import_module(
                'pyVHR.datasets.' + moduleName)
            classOBJ = getattr(datasetModule, className)
            obj = classOBJ(videodataDIR, BVPdataDIR)
        else:
            moduleName = datasetName.lower()
            className = datasetName.upper()
            relpath = str(path) + str(moduleName) + '.py'
            spec = util.spec_from_file_location(
                moduleName, relpath)
            mod = util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            classOBJ = getattr(mod, className)
            obj = classOBJ(videodataDIR, BVPdataDIR)

    except (AttributeError, ModuleNotFoundError):
        raise ImportError(
            '{} is not part of pyVHR dataset collection!'.format(datasetName))

    return obj


class Dataset(metaclass=ABCMeta):
    """
    This is the abstract class used for creating a new Dataset Class. 
    """

    def __init__(self, videodataDIR=None, BVPdataDIR=None):
        """
        Args:
            videodataDIR (str): path of the video data directory.
            BVPdataDIR (str): path of the ground truth BVP data directory.
        """
        # -- load filenames
        self.videoFilenames = []  # list of all video filenames
        self.sigFilenames = []    # list of all Sig filenames
        self.numVideos = 0        # num of videos in the dataset
        self.videodataDIR = videodataDIR
        self.BVPdataDIR = BVPdataDIR
        self.loadFilenames()

    def loadFilenames(self):
        """Load dataset file names: define vars videoFilenames and BVPFilenames."""

        # -- loop on the dir struct of the dataset getting filenames
        for root, dirs, files in os.walk(self.videodataDIR):
            for f in files:
                filename = os.path.join(root, f)
                path, name = os.path.split(filename)

                # -- select video
                if filename.endswith(self.video_EXT) and (name.find(self.VIDEO_SUBSTRING) >= 0):
                    self.videoFilenames.append(filename)

                # -- select signal
                if filename.endswith(self.SIG_EXT) and (name.find(self.SIG_SUBSTRING) >= 0):
                    self.sigFilenames.append(filename)

        # -- number of videos
        self.numVideos = len(self.videoFilenames)

    def getVideoFilename(self, videoIdx=0):
        """Get video file name given the progressive index."""
        return self.videoFilenames[videoIdx]

    def getSigFilename(self, videoIdx=0):
        """Get Signal file name given the progressive index."""
        return self.sigFilenames[videoIdx]

    @abstractmethod
    def readSigfile(self, filename):
        """ 
        Load signal from file.
        
        Returns:
            a :py:class:`pyVHR.BPM.BPM.BVPsignal` object that can be used to extract BPM signal from ground truth BVP signal.
        """
        pass



import numpy as np
import scipy.sparse
from numba import prange, jit
from scipy import stats
from scipy.signal import butter, filtfilt, savgol_filter


"""
This module contains a collection of filter methods.

FILTER METHOD SIGNATURE
A Filter method must accept theese parameters:
    > signal -> RGB signal as float32 ndarray with shape [num_estimators, rgb_channels, num_frames],
                or BVP signal as float32 ndarray with shape [num_estimators, num_frames].
    > **kargs [OPTIONAL] -> usefull parameters passed to the filter method.
It must return a filtered signal with the same shape as the input signal.
"""

def apply_filter(windowed_sig, filter_func, fps=None, params={}):
    """
    Apply a filter method to a windowed RGB signal or BVP signal. 

    Args:
        windowed_sig: list of length num_window of RGB signal as float32 ndarray with shape [num_estimators, rgb_channels, num_frames],
                      or BVP signal as float32 ndarray with shape [num_estimators, num_frames].
        filter_func: filter method that accept a 'windowed_sig' (pyVHR implements some filters in pyVHR.BVP.filters).
        params (dict): usefull parameters passed to the filter method.
    
    Returns:
        A filtered signal with the same shape as the input signal.
    """
    if 'fps' in params and params['fps'] == 'adaptive' and fps is not None:
        params['fps'] = np.float32(fps)
    filtered_windowed_sig = []
    for idx in range(len(windowed_sig)):
        transform = False
        sig = np.copy(windowed_sig[idx])
        if len(sig.shape) == 2:
            transform = True
            sig = np.expand_dims(sig, axis=1)
        if params == {}:
            filt_temp = filter_func(sig)
        else:
            filt_temp = filter_func(sig, **params)
        if transform:
            filt_temp = np.squeeze(filt_temp, axis=1)
        filtered_windowed_sig.append(filt_temp)
    return filtered_windowed_sig


# ------------------------------------------------------------------------------------- #
#                                     FILTER METHODS                                    #
# ------------------------------------------------------------------------------------- #


def BPfilter(sig, **kargs):
    """
    Band Pass filter (using BPM band) for RGB signal and BVP signal.

    The dictionary parameters are: {'minHz':float, 'maxHz':float, 'fps':float, 'order':int}
    """
    x = np.array(np.swapaxes(sig, 1, 2))
    b, a = butter(kargs['order'], Wn=[kargs['minHz'],
                                      kargs['maxHz']], fs=kargs['fps'], btype='bandpass')
    y = filtfilt(b, a, x, axis=1)
    y = np.swapaxes(y, 1, 2)
    return y


def zscore(sig):
    """
    Z-score filter for RGB signal and BVP signal.
    """
    x = np.array(np.swapaxes(sig, 1, 2))
    y = stats.zscore(x, axis=2)
    y = np.swapaxes(y, 1, 2)
    return y

 
def detrend(X, **kargs):
    """
    Detrending filter for RGB signal and BVP signal. 

    This code is based on the following article "An advanced detrending method with application
    to HRV analysis". Tarvainen et al., IEEE Trans on Biomedical Engineering, 2002.

    Code like: https://www.idiap.ch/software/bob/docs/bob/bob.rppg.base/v1.0.3/_modules/bob/rppg/cvpr14/filter_utils.html

    The dictionary parameters are: {'detLambda':int,}. Where 'detLambda' is the smoothing parameter.
    """
    if 'detLambda' not in kargs:
        kargs['detLambda'] = 10

    X = np.swapaxes(X, 1, 2)
    result = np.zeros_like(X)
    i = 0
    t = X.shape[1]
    l = t/kargs['detLambda']  # lambda
    I = np.identity(t)
    D2 = scipy.sparse.diags([1, -2, 1], [0, 1, 2], shape=(t-2, t)).toarray()
    Hinv = np.linalg.inv(I+l**2*(np.transpose(D2).dot(D2)))
    for est in X:
        detrendedX = (I - Hinv).dot(est)
        result[i] = detrendedX
        i += 1
    result = np.swapaxes(result, 1, 2)
    return result.astype(np.float32)

def zeromean(X):
    """
    Zero Mean filter for RGB signal and BVP signal. 
    """
    M = np.mean(X, axis=2)
    return X - np.expand_dims(M, axis=2) 

def sg_detrend(X, **kargs):
    """
    Remove the low-frequency components with the low-pass filter developed by Savitzky-Golay.
    It can be used for RGB signals and BVP signals.

    The dictionary parameters are: {'window_length':int, 'polyorder':int}. Where
    'window_length' is the length of the filter window (i.e. the number of coefficients). `window_length` must be a positive odd integer.
    'polyorder' is the order of the polynomial used to fit the samples. `polyorder` must be less than `window_length`.
    """
    if 'window_length' not in kargs:
        kargs['window_length'] = 31
    if 'polyorder' not in kargs:
        kargs['polyorder'] = 5

    trend = savgol_filter(X, window_length=kargs['window_length'], polyorder=kargs['polyorder'], axis=2)
    return X - trend


@jit(["int32[:](float32[:,:,:], int32, int32)", "int32[:](float64[:,:,:], int32, int32)", "int32[:](int32[:,:,:], int32, int32)"], nopython=True, nogil=True, parallel=True, fastmath=True)
def kernel_rgb_filter_th(sig, RGB_LOW_TH, RGB_HIGH_TH):
    """
    This method performs the Color Threshold filter for RGB signal only.
    Please refer to the method pyVHR.BVP.filters.rgb_filter_th.
    """
    goodidx = np.ones((sig.shape[0],), dtype=np.int32)
    for idx in prange(sig.shape[0]):
        b_in = 1
        for f in prange(sig.shape[2]):
            if ((sig[idx][0][f] <= RGB_LOW_TH and sig[idx][1][f] <= RGB_LOW_TH and sig[idx][2][f] <= RGB_LOW_TH) or
                    (sig[idx][0][f] >= RGB_HIGH_TH and sig[idx][1][f] >= RGB_HIGH_TH and sig[idx][2][f] >= RGB_HIGH_TH)):
                b_in = 0
        goodidx[idx] = b_in
    return goodidx


def rgb_filter_th(sig, **kargs):
    """
    Color Threshold filter for RGB signal only.

    The i_th estimators is filterd if its signal is outside the
    LOW/HIGH thresholds color interval in at least one frame.

    The dictionary parameters are: {'RGB_LOW_TH':int, 'RGB_HIGH_TH':int}.
    Where 'RGB_LOW_TH' and 'RGB_HIGH_TH' are RGB value ([0,255]) which describe the filter thresholds.
    """
    goodidx = kernel_rgb_filter_th(
        sig, np.int32(kargs['RGB_LOW_TH']), np.int32(kargs['RGB_HIGH_TH']))
    if np.sum(goodidx) == 0:
        return np.zeros((0, sig.shape[1], sig.shape[2]))
    return np.copy(sig[np.argwhere(goodidx).flatten()])



#!/usr/bin/python
# -*- encoding: utf-8 -*-

from logger import setup_logger
from model import BiSeNet
from face_dataset import FaceMask
from loss import OhemCELoss
from evaluate import evaluate
from optimizer import Optimizer
import cv2
import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.nn.functional as F
import torch.distributed as dist

import os
import os.path as osp
import logging
import time
import datetime
import argparse


respth = './res'
if not osp.exists(respth):
    os.makedirs(respth)
logger = logging.getLogger()


def parse_args():
    parse = argparse.ArgumentParser()
    parse.add_argument(
            '--local_rank',
            dest = 'local_rank',
            type = int,
            default = -1,
            )
    return parse.parse_args()


def train():
    args = parse_args()
    torch.cuda.set_device(args.local_rank)
    dist.init_process_group(
                backend = 'nccl',
                init_method = 'tcp://127.0.0.1:33241',
                world_size = torch.cuda.device_count(),
                rank=args.local_rank
                )
    setup_logger(respth)

    # dataset
    n_classes = 19
    n_img_per_gpu = 16
    n_workers = 8
    cropsize = [448, 448]
    data_root = '/home/zll/data/CelebAMask-HQ/'

    ds = FaceMask(data_root, cropsize=cropsize, mode='train')
    sampler = torch.utils.data.distributed.DistributedSampler(ds)
    dl = DataLoader(ds,
                    batch_size = n_img_per_gpu,
                    shuffle = False,
                    sampler = sampler,
                    num_workers = n_workers,
                    pin_memory = True,
                    drop_last = True)

    # model
    ignore_idx = -100
    net = BiSeNet(n_classes=n_classes)
    net.cuda()
    net.train()
    net = nn.parallel.DistributedDataParallel(net,
            device_ids = [args.local_rank, ],
            output_device = args.local_rank
            )
    score_thres = 0.7
    n_min = n_img_per_gpu * cropsize[0] * cropsize[1]//16
    LossP = OhemCELoss(thresh=score_thres, n_min=n_min, ignore_lb=ignore_idx)
    Loss2 = OhemCELoss(thresh=score_thres, n_min=n_min, ignore_lb=ignore_idx)
    Loss3 = OhemCELoss(thresh=score_thres, n_min=n_min, ignore_lb=ignore_idx)

    ## optimizer
    momentum = 0.9
    weight_decay = 5e-4
    lr_start = 1e-2
    max_iter = 80000
    power = 0.9
    warmup_steps = 1000
    warmup_start_lr = 1e-5
    optim = Optimizer(
            model = net.module,
            lr0 = lr_start,
            momentum = momentum,
            wd = weight_decay,
            warmup_steps = warmup_steps,
            warmup_start_lr = warmup_start_lr,
            max_iter = max_iter,
            power = power)

    ## train loop
    msg_iter = 50
    loss_avg = []
    st = glob_st = time.time()
    diter = iter(dl)
    epoch = 0
    for it in range(max_iter):
        try:
            im, lb = next(diter)
            if not im.size()[0] == n_img_per_gpu:
                raise StopIteration
        except StopIteration:
            epoch += 1
            sampler.set_epoch(epoch)
            diter = iter(dl)
            im, lb = next(diter)
        im = im.cuda()
        lb = lb.cuda()
        H, W = im.size()[2:]
        lb = torch.squeeze(lb, 1)

        optim.zero_grad()
        out, out16, out32 = net(im)
        lossp = LossP(out, lb)
        loss2 = Loss2(out16, lb)
        loss3 = Loss3(out32, lb)
        loss = lossp + loss2 + loss3
        loss.backward()
        optim.step()

        loss_avg.append(loss.item())

        #  print training log message
        if (it+1) % msg_iter == 0:
            loss_avg = sum(loss_avg) / len(loss_avg)
            lr = optim.lr
            ed = time.time()
            t_intv, glob_t_intv = ed - st, ed - glob_st
            eta = int((max_iter - it) * (glob_t_intv / it))
            eta = str(datetime.timedelta(seconds=eta))
            msg = ', '.join([
                    'it: {it}/{max_it}',
                    'lr: {lr:4f}',
                    'loss: {loss:.4f}',
                    'eta: {eta}',
                    'time: {time:.4f}',
                ]).format(
                    it = it+1,
                    max_it = max_iter,
                    lr = lr,
                    loss = loss_avg,
                    time = t_intv,
                    eta = eta
                )
            logger.info(msg)
            loss_avg = []
            st = ed
        if dist.get_rank() == 0:
            if (it+1) % 5000 == 0:
                state = net.module.state_dict() if hasattr(net, 'module') else net.state_dict()
                if dist.get_rank() == 0:
                    torch.save(state, './res/cp/{}_iter.pth'.format(it))
                evaluate(dspth='/home/zll/data/CelebAMask-HQ/test-img', cp='{}_iter.pth'.format(it))

    #  dump the final model
    save_pth = osp.join(respth, 'model_final_diss.pth')
    # net.cpu()
    state = net.module.state_dict() if hasattr(net, 'module') else net.state_dict()
    if dist.get_rank() == 0:
        torch.save(state, save_pth)
    logger.info('training done, model saved to: {}'.format(save_pth))


if __name__ == "__main__":
    train()



from sklearn.metrics import silhouette_score, pairwise_distances
from numpy import linalg as LA
from sklearn.cluster import KMeans
#from spherecluster import VonMisesFisherMixture
from sklearn.neighbors import KernelDensity
from lmfit import Model
from scipy.stats import iqr
import numpy as np
from scipy.signal import welch

"""
This module contains usefull methods used in pyVHR.BPM.BPM and
pyVHR.plot.visualize.
"""

def Welch(bvps, fps, minHz=0.65, maxHz=4.0, nfft=2048):
    """
    This function computes Welch'method for spectral density estimation.

    Args:
        bvps(flaot32 numpy.ndarray): BVP signal as float32 Numpy.ndarray with shape [num_estimators, num_frames].
        fps (float): frames per seconds.
        minHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        maxHz (float): frequency in Hz used to isolate a specific subband [minHz, maxHz] (esclusive).
        nfft (int): number of DFT points, specified as a positive integer.
    Returns:
        Sample frequencies as float32 numpy.ndarray, and Power spectral density or power spectrum as float32 numpy.ndarray.
    """
    _, n = bvps.shape
    if n < 256:
        seglength = n
        overlap = int(0.8*n)  # fixed overlapping
    else:
        seglength = 256
        overlap = 200
    # -- periodogram by Welch
    F, P = welch(bvps, nperseg=seglength, noverlap=overlap, fs=fps, nfft=nfft)
    F = F.astype(np.float32)
    P = P.astype(np.float32)
    # -- freq subband (0.65 Hz - 4.0 Hz)
    band = np.argwhere((F > minHz) & (F < maxHz)).flatten()
    Pfreqs = 60*F[band]
    Power = P[:, band]
    return Pfreqs, Power

def circle_clustering(W, eps=0.01, theta0=None, normalize=False):
    """
    TODO:documentare    

    Args:
        W (tipo): spiegare W
    
    Returns:
        cosa ritorna (tipo): spiegare , questa funzione restituisce: P, kjs, Q, hahdh
    """
    # general vars
    PI = np.pi
    n = W.shape[0]

    # param check
    if normalize:
        W = W / LA.norm(W)
    if theta0 is None:
        theta = 2*PI*np.random.rand(n)  # init. values in [0, 2*PI]
    else:
        theta = theta0

    # preliminar computations 
    sin_t = np.sin(theta)
    cos_t = np.cos(theta)
    A = np.dot(W, cos_t)
    B = np.dot(W, sin_t)

    # main loop
    ok = True
    rounds = 0
    while ok:
        ok = False
        rounds += 1
        nchanges = 0
        # loop on angles
        for i in range(n):
            old = theta[i]

            # change i-th theta
            #print(A[i], np.arctan(B[i]/A[i]))
            if np.abs(A[i])==0:
              if np.sign(A[i]*B[i])>0:
                theta[i] = PI/2
              else:
                theta[i] = -PI/2
            else:
              theta[i] = np.arctan(B[i]/A[i])    # within [-PI/2, PI/2]
            if A[i] >= 0:
                theta[i] += PI
            elif B[i] > 0:
                theta[i] += 2*PI

            # update Ak & Bk by elementwise product and diff
            A += np.multiply(W[i,:], np.repeat(np.cos(theta[i]) - np.cos(old), n))
            B += np.multiply(W[i,:], np.repeat(np.sin(theta[i]) - np.sin(old), n)) 

            if min(abs(old-theta[i]),abs(2*PI-old+theta[i])) > eps:
                ok = True
                nchanges += 1

    return theta

def optimize_partition(theta, out_fact=1):
    n = theta.shape[0]
    T = theta[:,None] - theta # x[:,None] adds a second axis to the array
    T = np.cos(T) - np.eye(n)
    OK = True

    # compute partitions P,Q
    P = []
    Q = []
    Tmean = np.mean(theta)
    for idx,th in enumerate(theta):
        if np.sign(np.sin(Tmean-th)) > 0:
            P.append(idx)
        else:
            Q.append(idx)

    # adjust partitions P,Q
    while OK:
        OK = False
        for i in range(n):
            if i in P:
                if np.sum(T[i,P]) < np.sum(T[i,Q]) and len(P) > 1:
                    Q.append(i)
                    P.remove(i)
                    OK = True
            else: 
                if np.sum(T[i,P]) > np.sum(T[i,Q]) and len(Q) > 1:
                    Q.remove(i)
                    P.append(i)
                    OK = True

    # pull out outliers from P and Q
    A = []
    B = []
    Z = []
    for i in P:
        A.append(np.sum(T[i,P]) / (T[i,P].shape[0]-1))

    M = np.max(A)
    M_idx = np.argmax(A)
    S = np.std(A)
    max_theta_P = theta[P[M_idx]] 
    L = M - S 
  
    # prune outliers in P
    for idx,i in enumerate(P):
        if M_idx != idx and A[idx] < L:
            Z.append(i)

    for i in Q:
        B.append(np.sum(T[i,Q]) / (T[i,Q].shape[0]-1))
    M = np.max(B)
    M_idx = np.argmax(B)
    S = np.std(B)
    max_theta_Q = theta[Q[M_idx]] 
    L = M - S  
  
    # prune outliers in Q
    for idx,i in enumerate(Q):
        if M_idx != idx and B[idx] < L:
            Z.append(i)

    P = list(set(P) - set(Z))
    Q = list(set(Q) - set(Z))

    return P, Q, Z, max_theta_P, max_theta_Q 

def gaussian(x,a,mu,sigma):
    return a*np.exp(-(x-mu)**2/(2*sigma**2))

def gaussian_fit(model,p, x, mu, max):
    result = model.fit(p, x=x, a=max, mu=mu, sigma=1)
    sigma = result.params['sigma'].value
    g = gaussian(x, max, mu, sigma)
    return result, g, sigma
  


#!/usr/bin/python
# -*- encoding: utf-8 -*-

import os.path as osp
import os
import cv2
from transform import *
from PIL import Image

face_data = '/home/zll/data/CelebAMask-HQ/CelebA-HQ-img'
face_sep_mask = '/home/zll/data/CelebAMask-HQ/CelebAMask-HQ-mask-anno'
mask_path = '/home/zll/data/CelebAMask-HQ/mask'
counter = 0
total = 0
for i in range(15):

    atts = ['skin', 'l_brow', 'r_brow', 'l_eye', 'r_eye', 'eye_g', 'l_ear', 'r_ear', 'ear_r',
            'nose', 'mouth', 'u_lip', 'l_lip', 'neck', 'neck_l', 'cloth', 'hair', 'hat']

    for j in range(i * 2000, (i + 1) * 2000):

        mask = np.zeros((512, 512))

        for l, att in enumerate(atts, 1):
            total += 1
            file_name = ''.join([str(j).rjust(5, '0'), '_', att, '.png'])
            path = osp.join(face_sep_mask, str(i), file_name)

            if os.path.exists(path):
                counter += 1
                sep_mask = np.array(Image.open(path).convert('P'))
                # print(np.unique(sep_mask))

                mask[sep_mask == 225] = l
        cv2.imwrite('{}/{}.png'.format(mask_path, j), mask)
        print(j)

print(counter, total)